#include "AlignSpectrometer.h"
#include <vector>
#include <iostream>
#include <math.h>
#include "Hit.h"
using namespace std;

AlignSpectrometer::AlignSpectrometer():  //constructor default values 
                               //RUN
                               numberOfEvents(10000),
                               beamMomentum(10.00),
                               thetaxz(0.0),
                               //GEOMETRY
                               spectrometerLength(30.),
                               integralBdL(0.0),
                               magLength(10.),
                               numberOfPlanes(6),
                               distBetweenPlanes(10.),
                               pixelSize(0.002),
                               resolution(0.0006),
                               ySize(2.),
                               zSize(1.),
                               tailAmplitude(0.1),
                               tailWidth(0.0018),
                               multScattAngle(0.00002),
                               //NOISE
                               noiseOccupancy(0.00001),
                               hitEfficiency(0.97),
                               Cut1(4.),
                               Cut2(4.),
                               Cut3(16.),
                               // STAT
                               numberOfReconstructedTracks(0),
                               nTotalHits(0),
                               nNoiseHitsOnTrack(0),
                               numberOfInefficientTracks(0),
                               numberOfRejectedTracks(0),
			       debug(true),
	         	       //ALIGNMENT
			       numiter(3),
			       alignsmear(0.01),
                               local_alignment(false)
  {

    // If you want to change some parameter, overwrite it here, so as to remember the default.

    // Some data members are derived from others:
    distBetweenPlanes= spectrometerLength/(numberOfPlanes-1.);

    // The hit efficiency depends on the noise occupancy:
    hitEfficiency=1.-0.06/fmaxf(log10(noiseOccupancy)+7.0,0.06);

    //output file for histograms
    outfile = new TFile("alignspectrometer.root", "RECREATE");

  // random shifts in y
  misalign_y[0] = 0.01;
  misalign_y[1] = -0.0044;
  misalign_y[2] = 0.0078;
  misalign_y[3] = -0.0004;
  misalign_y[4] = 0.0082;
  misalign_y[5] = -0.0006;
 // random shifts in z
  misalign_z[0] = -0.0090;
  misalign_z[1] = -0.0008;
  misalign_z[2] =  0.000;
  misalign_z[3] = -0.0042;
  misalign_z[4] =  0.014;
  misalign_z[5] = -0.0098;
  for(int i=0; i<numberOfPlanes; i++) {
    corr_y[i]=0.;
    corr_z[i]=0.;
  }
    //derivative and second derivative of chisquared wrt alignment

 D2Chi2 = new TMatrixD(12,12);
 DChi2 = new TMatrixD(12,1);


    // Book histograms
    h1 = new TH1F("h1","y0 residuals",100,-.005,.005);
    h2 = new TH1F("h2","z0 residuals",100,-.005,.005);
    h3 = new TH1F("h3","ty residuals",100,-.01,.01);
    h4 = new TH1F("h4","tz residuals",100,-.01,.01);
    h6 = new TH1F("h6","y0 pull",100,-10.,10.);
    h7 = new TH1F("h7","z0 pull",100,-10.,10.);
    h8 = new TH1F("h8"," ty pull ",100,-10.,10.);
    h9 = new TH1F("h9"," tz pull ",100,-10.,10.);
    h11 = new TH1F("h11"," z chisquared at plane 3 ",80,0.,24.);
    h12 = new TH1F("h12"," total chisquared ",80,0.,24.);
//book special histograms
  hresy[0] = new TH1F("h30"," y residual plane 0",100,-0.025,0.025);
  hresy[1] = new TH1F("h31"," y residual plane 1",100,-0.025,0.025);
  hresy[2] = new TH1F("h32"," y residual plane 2",100,-0.025,0.025);
  hresy[3] = new TH1F("h33"," y residual plane 3",100,-0.025,0.025);
  hresy[4] = new TH1F("h34"," y residual plane 4 ",100,-0.025,0.025);
  hresy[5] = new TH1F("h35"," y residual plane 5",100,-0.025,0.025);
  
  hresz[0] = new TH1F("h40"," z residual plane 0",100,-0.025,0.025);
  hresz[1] = new TH1F("h41"," z residual plane 1",100,-0.025,0.025);
  hresz[2] = new TH1F("h42"," z residual plane 2",100,-0.025,0.025);
  hresz[3] = new TH1F("h43"," z residual plane 3",100,-0.025,0.025);
  hresz[4] = new TH1F("h44"," z residual plane 4",100,-0.025,0.025);
  hresz[5] = new TH1F("h45"," z residual plane 5",100,-0.025,0.025);


  
}
void AlignSpectrometer::Reset(int iter) {


    D2Chi2->Zero();         //second derivative of total chisquared
    DChi2->Zero();          //first derivative of total chisquared
    for(int i=0;i<6;i++) {
      hresy[i]->Reset();    //residual distribution per plane
      hresz[i]->Reset();
      //clear hit counters per plane
      numHit[i]=0;
      firstHit[i]=0;
    }

  // Clear Hits store
    hitdata.clear();
  // Clear various counters
    numberOfReconstructedTracks=0;
    nTotalHits=                 0;
    nNoiseHitsOnTrack=          0;
    numberOfInefficientTracks=0;
    numberOfRejectedTracks=0;
}


TruthTrack AlignSpectrometer::SimulateEvent() {
    //==========================================================================================
    // Simulate the event
    //==========================================================================================

    TruthTrack t(0.,thetaxz,0.,0.,1./beamMomentum);

    // propagate track through the planes before the magnet
    PropagateStraight(0,t);
    float* par=t.GetPar();
    //par[0] =z0
    //par[1] =dz/dx
    //par[2] =y0
    //par[3] =dy/x

    //Trace through magnet (to a good approximation),
    float dtheta=0.003*integralBdL/beamMomentum;// bend angle in the magnet
    par[2]+=(par[3]*distBetweenPlanes/2.);      // move in y to the magnet center
    float ang=atan(par[3])+dtheta;
    par[2]+=(tan(ang)*distBetweenPlanes/2.);    // and then on to the next plane
    par[3]=tan(ang);                            // update angle
    par[0]+=(distBetweenPlanes*par[1]);         // in z all the way through magnet
    t.SetPar(par);

    //Propagate track through the rest of the planes after the magnet
    PropagateStraight(numberOfPlanes/2,t);

    return t;
}

void AlignSpectrometer::PropagateStraight(int firstplane, TruthTrack& t) { 
    //
    // propagate from plane to plane in a field free region
    //-----------------------------------------------------------------------

    float* par=t.GetPar();
    int nh=(int)hitdata.size();

    for(int j=firstplane; j<firstplane+numberOfPlanes/2; j++) {    

      if(debug) cout << " propagating track. Now at plane " << j << endl;
      numHit[j]=0;
      float y,z,r;

      if(j==firstplane) {
        y=par[2];                            // track impact in first plane
        z=par[0];                         
      }else{
      //extrapolate track to next plane
        par[2]+=(distBetweenPlanes*par[3]);
        par[0]+=(distBetweenPlanes*par[1]);
        y=par[2];                            // track impact in second plane
        z=par[0];                         
      }
       //add multiple scattering
        r=rdm.Gaus()/beamMomentum;
        par[3]+=(r*multScattAngle);           // update angle due to MS
        r=rdm.Gaus()/beamMomentum;
        par[1]+=(r*multScattAngle);           // update angle due to MS
        if(debug) cout << " track y z at plane " << j << " "
                       << " " << y << " "
                       << z << " angles " << par[3] << " " << par[1] << endl;

        r=rdm.Uniform();
        bool ishit = false; //has this plane been hit for the first time? 

        // If this gives a hit (allowing for inefficiency and tracks missing the chip)
        if(r<hitEfficiency && fabs(y)<ySize && fabs(z)<zSize) {
          float r1=rdm.Gaus();
          y+=r1*resolution;                    // then smear impact by resolution
          r1=rdm.Uniform();                    // sometimes give extra smear
          if(r1<tailAmplitude) {
            float r2=rdm.Gaus();
            y+=r2*tailWidth;                   // extra smear (resolution tail)
          }
          float r3=rdm.Gaus();
          z+=r3*resolution;                    // smear impact by resolution
          r3=rdm.Uniform();                    // do we need extra smear?
          if(r3<tailAmplitude) {
            float r2=rdm.Gaus();
            z+=r2*tailWidth;                   // extra smear (resolution tail)
          }

          Misalign(z,y,j); // Here we both displace the measurements
                           // and try to correct for it
          Hit h(z,y,j,false);
          hitdata.push_back(h);       // update vector of hits in data
          if(!ishit) firstHit[j]=nh;  // index of first hit in this plane
          ishit = true;
          t.AddHit(nh);               // update vector of hit indices on track
          if(debug) cout << " hit " << nh << " at plane " << j <<
	     	          " y " << y << " z " << z << " expect y z "
			 << par[2] << " " << par[0] << endl;
          nh++;                       //update total number of hits
          numHit[j]++;                //update number of hits in this plane
      }

      
      // add noise to 500x500 pixel area ( around real hit )
      int noise=rdm.Poisson(250000*noiseOccupancy);
      if(noise>0) {
        for(int k=0;k<noise;k++) {
          r=rdm.Uniform();                  // placement of noise hit
          float ynoise= y+(r-0.5)*(500.*pixelSize);
          r=rdm.Uniform();
          float znoise= z+(r-0.5)*(500.*pixelSize);
          Hit hnoise(znoise,ynoise,j,true);
          hitdata.push_back(hnoise);        // update vector of hits in data
          if(!ishit) firstHit[j]=nh;
          ishit=true;
          nh++;
          numHit[j]++;
        }
      }

    }                                       //end plane simulation loop
}
 void AlignSpectrometer::Misalign(float& z, float& y, int plane) {
   z+=misalign_z[plane]; 
   y+=misalign_y[plane];
   z-=corr_z[plane]; 
   y-=corr_y[plane]; 
 }

float AlignSpectrometer::ReconstructEvent() {
    // =======================================================================
    //Reconstruct one track 6 planes
    //Hits in all planes are mandatory.
    // =======================================================================
  
  float chi2min=10000000.;     // Initialize output
  bestTrack.SetChi2(chi2min);


  for(int i=0;i<numberOfPlanes;i++) {
    if(debug) cout << "  Hits in plane " << i
                   << " : " << numHit[i] << endl;
    if(debug) cout << " total hits " << hitdata.size() << endl;
    if(numHit[i]==0) return chi2min;
  }
    

    // loop over the hits in the first plane
    // =======================================================================
    for( int i0=firstHit[0];i0< firstHit[0]+numHit[0]; i0++) {
      if(chi2min < Cut3) continue;

      // loop over the hits in the second plane
      // =====================================================================
      for( int i1=firstHit[1];i1< firstHit[1]+numHit[1]; i1++) {

 	if(chi2min < Cut3) continue;
        double s2=resolution*resolution+alignsmear*alignsmear;

	//consider the xz plane - a non-bending plane
        //track state at plane 1
        TMatrixD z(2,1);
        z[0][0]= hitdata.at(i1).GetZ();
        z[1][0]= (hitdata.at(i1).GetZ()-hitdata.at(i0).GetZ())/distBetweenPlanes ;
	//its covariance
        TMatrixD Cz(2,2);
        Cz[0][0]=s2;
        Cz[0][1]=s2/distBetweenPlanes;
        Cz[1][0]=Cz[0][1];
        Cz[1][1]=2*s2/distBetweenPlanes/distBetweenPlanes;
        //its hits
        vector<int> hits;
        hits.push_back(i0);
        hits.push_back(i1);
        if(debug) cout << "  Seed track hits: " << i0 << " " << i1 << endl;
	//store it all in two separate track copies
        RecoTrack t2(z,Cz,hits);
        RecoTrack t3(z,Cz,hits);

        // loop over hits in the third plane
        //===========================================================
        for( int i2=firstHit[2];i2< firstHit[2]+numHit[2]; i2++) {

	  float chi2 = KalmanFilter(2,i2,t2,t3);  // t3 is the updated track
          if(debug) cout << " Hit in third plane, i2,chi2  " << i2 << " " << chi2 << endl;

 	  if(chi2 > Cut1 || chi2min < Cut3) continue;
 
          //loop over hits in the fourth plane
          //====================================================================================
          for( int i3=firstHit[3];i3< firstHit[3]+numHit[3]; i3++) {
            RecoTrack t4(t3.GetPar(),t3.GetCov(),t3.GetHits());
	    chi2 = KalmanFilter(3,i3,t3,t4);      //t4 is the updated track
            if(debug) cout << " Hit in fourth plane, i3,chi2  " << i3 << " " << chi2  << endl;
            h11->Fill(chi2);

 	    if(chi2 > Cut1 || chi2min < Cut3) continue;

            //loop over hits in the fifth plane
            //====================================================================================

            for( int i4=firstHit[4];i4< firstHit[4]+numHit[4]; i4++) {
              RecoTrack t5(t4.GetPar(),t4.GetCov(),t4.GetHits());
	      chi2 = KalmanFilter(4,i4,t4,t5); //t5 is the updated track (only if chi2 < Cut1)
              if(debug) cout << " Hit in fifth plane, i4,chi2  " << i4 << " " << chi2  << endl;

 	      if(chi2 > Cut1 || chi2min < Cut3) continue;

              //loop over hits in the sixth plane
              //====================================================================================
              for( int i5=firstHit[5];i5< firstHit[5]+numHit[5]; i5++) {
                RecoTrack t6(t5.GetPar(),t5.GetCov(),t5.GetHits());
  	        chi2 = KalmanFilter(5,i5,t5,t6);
                if(debug) cout << " Hit in sixth plane, i5,chi2  " << i5 << " " << chi2  << endl;
 	        if(chi2 > Cut1 || chi2min < Cut3) continue;

	        // now we have a track candidate for the six-plane configuration

	        // make a global chi2 fit and store it if chi2 is small enough
                // ===============================================================================
                TMatrixD x(4,1);
                TMatrixD C(4,4);
                hits=t6.GetHits();         // get hit list
                RecoTrack tg(x,C,hits);
                chi2= GlobalChi2(tg);
                if(debug) cout << " Global fit chi2  " << chi2 << endl;
                if(chi2<Cut3) {
		// store it
		  bestTrack.SetPar(tg.GetPar()); 
		  bestTrack.SetCov(tg.GetCov()); 
		  bestTrack.SetHits(tg.GetHits()); 
		  bestTrack.SetChi2(tg.GetChi2()); 
		  bestTrack.SetNdf(tg.GetNdf()); 
                  chi2min=chi2;
		}
 	      }
	    }
	  }
	}
      }
    }
    return chi2min;
}


float AlignSpectrometer::KalmanFilter(int p1, int ihit, RecoTrack t, RecoTrack& tout) {
  // Propagates a track candidate from detector plane p1-1 to detector plane p1
  // as a straight line in x-z (the non-bending plane)
  // updates the track parameters and their error matrix in x-z
  // returns the chisquared at detector plane p1 for hit number ihit in this plane

        if(debug) cout << " in Kalman Filter " << endl; 
        double s2=resolution*resolution+alignsmear*alignsmear;
        double pinv = 1./beamMomentum; //use here a fixed momentum estimate
  	double t0=multScattAngle*pinv; //average multiple scattering angle
        TMatrixD z=t.GetPar();         // current track parameters
        TMatrixD C=t.GetCov();         // current covariance matrix
        vector<int> hits=t.GetHits();  // get hit list
        int plane0= hitdata.at(hits.back()).GetPlane();
        double d=distBetweenPlanes*(p1-plane0); // propagation distance
        if(debug) cout << "   propagating from plane " << plane0 << " to " << p1 << " dist " <<  d << endl; 
        TMatrixD Fz(2,2);  //propagator Fz to next plane
        TMatrixD FTz(2,2); //transposed Fz matrix
        Fz[0][0]=1;     
        Fz[0][1]=d;
        Fz[1][0]=0;     
        Fz[1][1]=1;

        FTz=Fz.T();  //the transpose of F
        Fz.T();

        if(debug) cout << "   adding mult scatt " << endl; 
        //multiple scattering contribution to covariance of extrapolation
        TMatrixD Qz(2,2);
        Qz[0][0]=t0*t0*d*d;
        Qz[0][1]=t0*t0*d;
        Qz[1][0]=t0*t0*d;
        Qz[1][1]=t0*t0;

 	//covariance of extrapolation
        TMatrixD Cpz(2,2);
        Cpz = Fz*C*FTz+Qz;

        //predicted state at plane p1
        TMatrixD zpred(2,1);
        zpred = Fz*z;

        //covariance matrix of updated state 
        TMatrixD Cinv(2,2); //inverse covariance matrix of updated state
        TMatrixD Minv(2,2); //inverse covariance matrix of the new measurement
        Minv.Zero();
        Minv[0][0]=1./s2;
        Cinv = Cpz.Invert() + Minv; //add inverted covariances
        C=Cinv.Invert();            //update covariance

        //update track with new measurement
        TMatrixD znew(2,1);       //new measurement weighted by 1/s2
        znew[0][0]=hitdata.at(ihit).GetZ()/s2;
        znew[1][0]=0.;
        z=C*(Cpz*zpred + znew);  //update track params - remember Cpz is inverted
        Cpz.Invert();


        //The residual of the new measurement wrt the prediction 
        double r =(hitdata.at(ihit).GetZ()-zpred[0][0]);
        double R = s2 + Cpz[0][0];
        float chi2=r*r/R;

        if(chi2<Cut1) {            //update track
          hits.push_back(ihit);    //update hit list
          tout.SetPar(z); 
          tout.SetCov(C);
          tout.SetHits(hits);
          tout.SetChi2(chi2);
          tout.SetNdf(7);
	}
 
        return chi2;
}
//**********************************************************************
float AlignSpectrometer::GlobalChi2(RecoTrack& t) {
  //global chi2-fit to x-y hits in a number of pixel planes

  TMatrixD x(4,1);
  x=t.GetPar();
  TMatrixD C(4,4);
  C=t.GetCov();
  float pinv=1./beamMomentum;
  double t0=multScattAngle*pinv; // MS angle using an estimated 1/p
  double t2=t0*t0;
  double d2=distBetweenPlanes*distBetweenPlanes;

  double s2=resolution*resolution+alignsmear*alignsmear;
  vector<int> hits=t.GetHits();

  //
  // The column vector of the measurements
  TMatrixD m(12,1);
  if(debug) cout << " Global fit to " << hits.size() << " hits " << endl;
  for (int i=0 ;i<6;i++) {
    Hit h=hitdata.at(hits.at(i));
    m[i][0]=h.GetZ();
    m[i+6][0]=h.GetY();
    if(debug) cout << "     Hit " << i << " z " <<  m[i][0] << " y " << m[i+6][0] << endl;
  }

  //
  // The Covariance matrix of the measurements, including multiple scattering

  TMatrixD V(12,12);
  TMatrixD V0(12,12);
  V.Zero();
  V0.Zero();
  V[0][0]=s2;
  V[6][6]=s2;
  V0[0][0]=s2;
  V0[6][6]=s2;
  for(int i=1; i<6; i++) {
    for(int j=i; j<6; j++) {
      V[i][j] = V[i-1][j-1] + i*j*t2*d2;
      V[i+6][j+6] = V[i][j];
      V0[i][j]=V0[i-1][j-1];
      V0[i+6][j+6]=V0[i-1][j-1];

       if(j>i) {
         V[j][i] = V[i][j];
         V[j+6][i+6] = V[i+6][j+6];
         V0[j][i] = V0[i][j];
         V0[j+6][i+6] = V0[i+6][j+6];
       }
       if(debug) cout << " V: " << i << " " << j << " " << V[i][j] << " " << V[i+6][j+6] << endl;
    }
  }

  //
  // The track state vector is defined as
  // x = (z0, z', y0, y') with
  // track impact z0, y0 and slopes z'=dz/dx and y'=dy/dx all given at plane 0
  //
  // H projects the track state vector on the measurement base
 TMatrixD H(12,4);
 H.Zero();
  for(int i=0; i<6; i++) {
    Hit h=hitdata.at(hits.at(i));
    int plane=h.GetPlane();
    int j=i+6;
    H[i][0]=1;
    H[j][2]=1;
    H[i][1]=plane*distBetweenPlanes;
    H[j][3]=plane*distBetweenPlanes;
    if(debug) cout << " H: " << i << " " << H[i][0] << " " << H[j][2] << " " << H[i][1] << " " << H[j][3] << endl;
  }

  TMatrixD HT=H.T(); //The transposed
  H.T();             //Do this to transpose H back.
  //
   //
  C=HT*V.Invert()*H;
  C.Invert();
  // V is now inverse
  // C is now the covariance matrix of the fitted track state
  // x is the fitted track state vector
  x = C*HT*V*m;

  // R is the residual vector
  TMatrixD R=m-H*x;
  TMatrixD RT=R.T();
  R.T();          //Do this because R got also transposed
  TMatrixD Chi2=RT*V*R;
  
  if(Chi2[0][0]<Cut3) {
    t.SetPar(x);             //update track
    t.SetCov(C);
    t.SetChi2(Chi2[0][0]);
    t.SetNdf(8.);
    //accumulate local residuals
    for(int i=0; i<6; i++) {
      hresz[i]->Fill(R[i][0]);
      hresy[i]->Fill(R[i+6][0]);
    }

    //accumulate matrices for global alignment

    TMatrixD RC(12,12);    

    RC=V0-H*C*HT;  // Covariance of the residuals. Must use diagonal V0 here??
    RC.Invert();

    //Note that dR_y(z)/d(Delta_y(z)) ==1
    (*DChi2)+=(RC*R);
    (*D2Chi2)+=RC;

  }
  return Chi2[0][0];  
}
//*****************************************************************************************************
void AlignSpectrometer::PrintEvent(int event){


  TMatrixD par= bestTrack.GetPar();
  TMatrixD cov= bestTrack.GetCov();
  float chi2= bestTrack.GetChi2();
  int ndf= bestTrack.GetNdf();
  vector<int> hits= bestTrack.GetHits();

  h12->Fill(chi2);

  //Print summary of this event
  if(chi2 < Cut3) {

    numberOfReconstructedTracks++;
    if(debug) cout << " " << endl;
    if(debug) cout << " New reconstructed track: " << endl;
    if(debug) cout << " z0 " << par[0][0] << " +- " << sqrt(cov[0][0]) << endl;
    if(debug) cout << " dz/dx " << par[1][0] << " +- " << sqrt(cov[1][1])  << endl;
    if(debug) cout << " y0 " << par[2][0]  << " +- " << sqrt(cov[2][2]) << endl;
    if(debug) cout << " dy/dx " << par[3][0]  << " +- " << sqrt(cov[2][2]) << endl;
    if(debug) cout << " Chi2 " << chi2  << "  for ndf " << ndf << endl;

    //Fill histograms
    h1->Fill( par[2][0] );
    h2->Fill( par[0][0] );
    h3->Fill( par[3][0] );
    h4->Fill( par[1][0]-thetaxz );
    h6->Fill( par[2][0]/sqrt(cov[2][2]) );
    h7->Fill( par[0][0]/sqrt(cov[0][0]) );
    h8->Fill( par[3][0]/sqrt(cov[3][3]) );
    h9->Fill( (par[1][0]-thetaxz)/sqrt(cov[1][1]) );
    int nh=(int)hits.size();
    if(nh<numberOfPlanes) numberOfInefficientTracks++;
    for(int i=0; i<nh; i++) {
      nTotalHits++;
      if(hitdata.at(hits.at(i)).IsNoise()) nNoiseHitsOnTrack++;
    }
    if(debug) cout << " number of hits " << nh << endl;

  } else {
    if(debug) cout << " fit failed  chisquared=" << chi2  << " ndf= " << ndf << endl;
    numberOfRejectedTracks++;
  }

  if(event>1) debug=false;
  if(event%1000==0) cout << " event " << event << endl;

  if(event==numberOfEvents-1) {
    cout << " " << endl;
    cout << " Generated Tracks " << numberOfEvents << endl;
    cout << " Reconstructed Tracks " << numberOfReconstructedTracks << endl;
    cout << " Tracks with missing hits " << numberOfInefficientTracks << endl;
    cout << " Tracks lost to quality cuts " << numberOfRejectedTracks << endl;
    cout << " Total hits on track " << nTotalHits << endl;
    cout << " Total noise hits on track " << nNoiseHitsOnTrack << endl;

  }

}
//*********************************************************************************************
 void AlignSpectrometer::Align() {
  //Update alignment corrections

   cout << " " << endl;
   cout << " New iteration " << endl;
   cout << " " << endl;

    TMatrixD param(12,1);

    D2Chi2->Invert();
    param=(*D2Chi2)*(*DChi2); // fitted alignment corrections


    for(int i=0; i<6; i++) {
      if(local_alignment) {
        corr_z[i]+=hresz[i]->GetMean();
        corr_y[i]+=hresy[i]->GetMean();
        (*D2Chi2)[i][i]=hresz[i]->GetRMS()*hresz[i]->GetRMS();
        (*D2Chi2)[i+6][i+6] =
             hresy[i]->GetRMS()*hresy[i]->GetRMS();
      } else {
        corr_z[i]+=param[i][0];
        corr_y[i]+=param[i+6][0];
      }
    }
    cout << endl; 
    cout << " Alignment corrections in z: " << endl; 
    for(int i=0; i<6; i++) {
     cout << " plane " << i << " Original Misalignment " << misalign_z[i] <<
	" Correction " << corr_z[i] << " Alignment error " << 
        misalign_z[i]-corr_z[i] << " +- " << sqrt((*D2Chi2)[i][i]) << endl;
    }

    cout << endl; 
    cout << " Alignment corrections in y: " << endl; 
    for(int i=0; i<6; i++) { 
      cout << " plane " << i << " Original Misalignment " << misalign_y[i] <<
        " Correction " << corr_y[i] << " Alignment error " <<
        misalign_y[i]-corr_y[i] <<  " +- " << sqrt((*D2Chi2)[i+6][i+6]) << endl; 
    }
        
    outfile->Write();    
 }





