#ifndef alignspectrometer_h
#define alignspectrometer_h
#include <vector>
#include "TruthTrack.h"
#include "RecoTrack.h"
#include "Hit.h"
#include <TMatrixD.h>
#include <TH1F.h>
#include <TFile.h>
#include <TRandom3.h>

// Implements a field-off alignment run for a 6-plane spectrometer
class AlignSpectrometer {
 public:
  AlignSpectrometer();
  void Reset(int iter);
  TruthTrack SimulateEvent();
  void PropagateStraight(int firstPlane, TruthTrack& nextTrack);
  float ReconstructEvent();
  float KalmanFilter(int p1, int ihit, RecoTrack t, RecoTrack& tout );
  float GlobalChi2(RecoTrack& t);
  void PrintEvent(int event);
  int GetNumberOfEvents() {return numberOfEvents;};
  int GetNumberOfIterations() {return numiter;};
  void Misalign(float& z, float&y, int plane);
  void Align();
  
 protected:

  //RUN
  const int numberOfEvents;        //number of events to be simulated   
  const float beamMomentum;        // GeV
  const float thetaxz;           // incident track angle in the xz plane

  //GEOMETRY
  const float spectrometerLength;  //(cm)
  const float integralBdL;         // B-field integral Tesla*cm
  const float magLength;           // cm
  const int numberOfPlanes;        //number of tracking planes
  float distBetweenPlanes;         //distance between planes
  //(assume half that on each side of magnet)
  const float pixelSize;          //(cm)
  const float resolution;         //measurement resolution (cm)
  const float ySize;              // Half height of chip
  const float zSize;              // Half width of chip

  const float tailAmplitude;       //probability of badly measured hit
  const float tailWidth;           //resolution of badly measured hit (cm)
// The multiple scattering times momentum per plane is estimated as follows
// Rl(50mu Si) = 5.3e-4, Rl(50cm He) = 8.8e-5, total Rl=5.5e-4
// multScattAngle=0.0175*sqrt(Rl)*(1+0.125*log(10Rl)/log(10))/sqrt(2)
  const float multScattAngle;    // effective theta0*E(GeV) (mult scatt) per plane

  //NOISE
  // There is an adjustable threshold with which we can get the noise occupancy
  // as low is 10^-7, at a cost in the hit efficiency.
  // Assume: (log10(noise)+7)*(1.-efficiency)=0.06
  const float noiseOccupancy;  // noise probability in readout time window
  float hitEfficiency;     // probability for a track to make a hit
  const float Cut1;        // cut on chisquared for first 3 hits
  const float Cut2;        // cut on chisquared for next hits
  const float Cut3;        // cut on total chisquared/ndf
  bool debug;              // debug flag

//ALIGNMENT
//introduce some initial misalignment. Here just some shifts in y-z (cm) of each plane.
  const int numiter;         //number of alignment iterations
  const double alignsmear;    //for inflation of errors during alignment
 //alignment method: choose local or global alignment method 
 const bool local_alignment;
 // random shifts in y
 float misalign_y[6];
 // random shifts in z
 float misalign_z[6];
 //alignment corrections
 float corr_y[6];
 float corr_z[6];
 //summed second and first derivatives of ch2 wrt alignment
 TMatrixD* D2Chi2; 
 TMatrixD* DChi2; 

  //RAW DATA
  std::vector<Hit> hitdata;
  int firstHit[6];
  int numHit[6];

  //RECONSTRUCTED DATA
  RecoTrack bestTrack;

  //COUNTERS
  int numberOfReconstructedTracks, numberOfGoodReconstructedTracks, nTotalHits,
    nNoiseHitsOnTrack, numberOfInefficientTracks, numberOfRejectedTracks,
    numberOfGoodRejectedTracks, iteration;

  // HISTOGRAMS
  TFile* outfile;
  TH1F* h1;
  TH1F* h2;
  TH1F* h3;
  TH1F* h4;
  TH1F* h5;
  TH1F* h6;
  TH1F* h7;
  TH1F* h8;
  TH1F* hresy[6];
  TH1F* hresz[6];
  TH1F* hz;
  TH1F* hy;
  TRandom3 rdm;

};
#endif

