#ifndef Spectrometer_h
#define Spectrometer_h
#include <vector>
#include "TruthTrack.h"
#include "RecoTrack.h"
#include "Hit.h"
#include "TMatrixD.h"
#include "TH1F.h"
#include "TFile.h"
#include "TRandom3.h"


class Spectrometer {
 public:
  Spectrometer();
  TruthTrack SimulateEvent();
  void PropagateStraight(int firstPlane, TruthTrack& nextTrack);
  float ReconstructEvent();
  float KalmanFilter(int p1, int ihit, RecoTrack t, RecoTrack& tout );
  float GlobalChi2(RecoTrack& t);
  void PrintEvent(int event);
  int GetNumberOfEvents() {return numberOfEvents;};
  
 private:
  //RUN
  const int numberOfEvents;        //number of events to be simulated   
  float beamMomentum;              // GeV
  const float thetaxz;            // incident track angle in the xz plane

  //GEOMETRY
  float spectrometerLength;       //(cm)
  float integralBdL;              // B-field integral Tesla*cm
  const float magLength;          // cm
  int numberOfPlanes;             //number of tracking planes
  float distBetweenPlanes;        //distance between planes
  //(assume half that on each side of magnet)
  const float pixelSize;          //(cm)
  const float resolution;         //measurement resolution (cm)
  const float ySize;              // Half height of chip
  const float zSize;              // Half width of chip

  const float tailAmplitude;       //probability of badly measured hit
  const float tailWidth;           //resolution of badly measured hit (cm)
// The multiple scattering times momentum per plane is estimated as follows
// Rl(50mu Si) = 5.3e-4, Rl(50cm He) = 8.8e-5, total Rl=5.5e-4
// multScattAngle=0.0175*sqrt(Rl)*(1+0.125*log(10Rl)/log(10))/sqrt(2)
  const float multScattAngle;    // effective theta0*E(GeV) (mult scatt) per plane

  //NOISE
  // There is an adjustable threshold with which we can get the noise occupancy
  // as low is 10^-7, at a cost in the hit efficiency.
  // Assume: (log10(noise)+7)*(1.-efficiency)=0.06
  float noiseOccupancy;    // noise probability in readout time window
  float hitEfficiency;     // probability for a track to make a hit
  float Cut1;              // cut on chisquared for first 3 hits
  float Cut2;              // cut on chisquared for next hits
  float Cut3;              // cut on total chisquared/ndf
  bool debug;              // debug flag

  //RAW DATA
  std::vector<Hit> hitdata;
  int firstHit[6];
  int numHit[6];

  //RECONSTRUCTED DATA
  RecoTrack bestTrack;

  //COUNTERS
  int numberOfReconstructedTracks, numberOfGoodReconstructedTracks, nTotalHits,
    nNoiseHitsOnTrack, numberOfInefficientTracks, numberOfRejectedTracks,
    numberOfGoodRejectedTracks;

  // HISTOGRAMS
  TFile* outfile;
  TH1F* h1;
  TH1F* h2;
  TH1F* h3;
  TH1F* h4;
  TH1F* h5;
  TH1F* h6;
  TH1F* h7;
  TH1F* h8;
  TH1F* h9;
  TH1F* h10;
  TH1F* h11;
  TH1F* h12;
  TRandom3 rdm;
};
#endif

